@@ -46,10 +46,11 @@ extern "C" {
 void _emval_register_symbol(const char*);
 
 enum {
-  _EMVAL_UNDEFINED = 1,
-  _EMVAL_NULL = 2,
-  _EMVAL_TRUE = 3,
-  _EMVAL_FALSE = 4
+  _EMVAL_UNDEFINED = 2,
+  _EMVAL_NULL = 4,
+  _EMVAL_TRUE = 6,
+  _EMVAL_FALSE = 8,
+  _EMVAL_LAST_RESERVED_HANDLE = 8,
 };
 
 typedef struct _EM_DESTRUCTORS* EM_DESTRUCTORS;
@@ -385,11 +386,13 @@ public:
   }
 
   val(const val& v) : val(v.as_handle()) {
-    internal::_emval_incref(handle);
+    if (uses_ref_count()) {
+      internal::_emval_incref(handle);
+    }
   }
 
   ~val() {
-    if (handle) {
+    if (uses_ref_count()) {
       internal::_emval_decref(as_handle());
       handle = 0;
     }
@@ -402,6 +405,13 @@ public:
     return handle;
   }
 
+  // Takes ownership of the handle away from, and invalidates, this instance.
+  EM_VAL release_ownership() {
+    EM_VAL taken = as_handle();
+    handle = 0;
+    return taken;
+  }
+
   val& operator=(val&& v) & {
     val tmp(std::move(v));
     this->~val();
@@ -540,6 +550,28 @@ public:
     return fromGenericWireType<T>(result);
   }
 
+#ifdef __wasm64__
+  template<>
+  long as<long>() const {
+    using namespace internal;
+
+    typedef BindingType<long> BT;
+    typename WithPolicies<>::template ArgTypeList<long> targetType;
+
+    return _emval_as_int64(as_handle(), targetType.getTypes()[0]);
+  }
+
+  template<>
+  unsigned long as<unsigned long>() const {
+    using namespace internal;
+
+    typedef BindingType<unsigned long> BT;
+    typename WithPolicies<>::template ArgTypeList<unsigned long> targetType;
+
+    return _emval_as_uint64(as_handle(), targetType.getTypes()[0]);
+  }
+#endif
+
   template<>
   int64_t as<int64_t>() const {
     using namespace internal;
@@ -608,6 +640,12 @@ private:
       : handle(handle), thread(pthread_self())
   {}
 
+  // Whether this value is a uses incref/decref (true) or is a special reserved
+  // value (false).
+  bool uses_ref_count() const {
+    return handle > reinterpret_cast<EM_VAL>(internal::_EMVAL_LAST_RESERVED_HANDLE);
+  }
+
   template<typename WrapperType>
   friend val internal::wrapped_extend(const std::string& , const val& );
 
@@ -764,9 +802,20 @@ template<typename T>
 struct BindingType<T, typename std::enable_if<std::is_base_of<val, T>::value &&
                                               !std::is_const<T>::value>::type> {
   typedef EM_VAL WireType;
+
+  // Marshall to JS with move semantics when we can invalidate the temporary val
+  // object.
+  static WireType toWireType(val&& v) {
+    return v.release_ownership();
+  }
+
+  // Marshal to JS with copy semantics when we cannot transfer the val objects
+  // reference count.
   static WireType toWireType(const val& v) {
     EM_VAL handle = v.as_handle();
-    _emval_incref(handle);
+    if (v.uses_ref_count()) {
+      _emval_incref(handle);
+    }
     return handle;
   }
   static T fromWireType(WireType v) {
