@@ -44,7 +44,7 @@ time_t _mktime_js(struct tm* tm);
 void _localtime_js(time_t t, struct tm* __restrict__ tm);
 void _gmtime_js(time_t t, struct tm* __restrict__ tm);
 
-void _tzset_js(long* timezone, int* daylight, char** tzname);
+void _tzset_js(long* timezone, int* daylight, char* std_name, char* dst_name);
 
 const char* emscripten_pc_get_function(uintptr_t pc);
 const char* emscripten_pc_get_file(uintptr_t pc);
@@ -146,6 +146,10 @@ void __resumeException(void* exn);
 void __cxa_call_unexpected(void* exn);
 void llvm_eh_typeid_for(void* exn);
 
+uint32_t _emscripten_lookup_name(const char *name);
+
+int _emscripten_system(const char *command);
+
 #ifdef __cplusplus
 }
 #endif
