@@ -11,7 +11,7 @@ extern "C" {
 int noted = 0;
 
 char* EMSCRIPTEN_KEEPALIVE note(int n) {
-  EM_ASM({ Module.noted = Number($0); out("set noted " + Module.noted) }, (long)&noted);
+  EM_ASM({ Module.noted = Number($0); out("set noted " + Module.noted) }, &noted);
   EM_ASM({ out([$0, $1]) }, n, noted);
   noted += n;
   EM_ASM({ out(['noted is now', $0]) }, noted);
