@@ -52,8 +52,8 @@ def clear(ports, settings, shared):
 
 
 def process_args(ports):
-  return ['-I' + ports.get_include_dir('bullet')]
+  return ['-isystem', ports.get_include_dir('bullet')]
 
 
 def show():
-  return 'bullet (USE_BULLET=1; zlib license)'
+  return 'bullet (-sUSE_BULLET=1 or --use-port=bullet; zlib license)'
